import 'dart:async';

import 'package:meta/meta.dart';
// ignore: test_library_import
import 'package:sodium/sodium.dart';
import 'package:test/test.dart' as t;

import 'cases/aead_test_case.dart';
import 'cases/auth_test_case.dart';
import 'cases/box_test_case.dart';
import 'cases/generic_hash_test_case.dart';
import 'cases/kdf_test_case.dart';
import 'cases/kx_test_case.dart';
import 'cases/pwhash_test_case.dart';
import 'cases/randombytes_test_case.dart';
import 'cases/secret_box_test_case.dart';
import 'cases/secret_stream_test_case.dart';
import 'cases/short_hash_test_case.dart';
import 'cases/sign_test_case.dart';
import 'cases/sodium_init_test_case.dart';
import 'cases/sodium_test_case.dart';
import 'test_case.dart';

typedef SetupFn = void Function(dynamic Function() body);
typedef GroupFn = void Function(String description, dynamic Function() body);
typedef TestFn = void Function(
  String description,
  dynamic Function() body, {
  bool isSumo,
});

abstract class TestRunner {
  late final Sodium _sodium;

  final bool isSumoTest;

  bool get is32Bit => false;

  TestRunner({
    required this.isSumoTest,
  });

  Iterable<TestCase> createTestCases() => [
        SodiumTestCase(this),
        SodiumInitTestCase(this),
        RandombytesTestCase(this),
        SecretBoxTestCase(this),
        SecretStreamTestCase(this),
        AeadTestCase(this),
        AuthTestCase(this),
        BoxTestCase(this),
        SignTestCase(this),
        GenericHashTestCase(this),
        ShortHashTestCase(this),
        PwhashTestCase(this, is32Bit: is32Bit),
        KdfTestCase(this),
        KxTestCase(this),
      ];

  @protected
  @visibleForTesting
  Future<Sodium> loadSodium();

  Sodium get sodium => _sodium;

  @visibleForOverriding
  late SetupFn setUpAll = t.setUpAll;

  @visibleForOverriding
  void test(
    String description,
    dynamic Function() body, {
    bool isSumo = false,
  }) =>
      t.test(
        description,
        body,
        skip: !isSumoTest && isSumo
            ? 'This test only works with the sodium.js sumo variant'
            : null,
      );

  @visibleForOverriding
  late GroupFn group = t.group;

  void setupTests() {
    final testCases = createTestCases().toList();

    setUpAll(() async {
      _sodium = await loadSodium();

      // ignore: avoid_print
      print(
        'Running integration tests with libsodium version: ${_sodium.version} '
        '(${is32Bit ? '32 bit' : '64 bit'})',
      );
    });

    for (final testCase in testCases) {
      group(testCase.name, () {
        testCase.setupTests();
      });
    }
  }
}
